@@ -415,7 +415,7 @@ g_ir_node_get_size (GIrNode *node)
       break;
 
     case G_IR_NODE_FUNCTION:
-      size = 16; 
+      size = sizeof (FunctionBlob);
       break;
 
     case G_IR_NODE_PARAM:
@@ -469,7 +469,7 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeStruct *struct_ = (GIrNodeStruct *)node;
 
-	size = 20;
+	size = 24;
 	for (l = struct_->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
       }
@@ -479,7 +479,7 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeBoxed *boxed = (GIrNodeBoxed *)node;
 
-	size = 20;
+	size = 24;
 	for (l = boxed->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
       }
@@ -517,7 +517,7 @@ g_ir_node_get_size (GIrNode *node)
       {
 	GIrNodeUnion *union_ = (GIrNodeUnion *)node;
 
-	size = 28;
+	size = 32;
 	for (l = union_->members; l; l = l->next)
 	  size += g_ir_node_get_size ((GIrNode *)l->data);
 	for (l = union_->discriminators; l; l = l->next)
@@ -707,7 +707,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeStruct *struct_ = (GIrNodeStruct *)node;
 
-	size = 20;
+	size = 24;
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (struct_->gtype_name)
 	  size += ALIGN_VALUE (strlen (struct_->gtype_name) + 1, 4);
@@ -722,7 +722,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeBoxed *boxed = (GIrNodeBoxed *)node;
 
-	size = 20;
+	size = 24;
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (boxed->gtype_name)
 	  {
@@ -814,7 +814,7 @@ g_ir_node_get_full_size_internal (GIrNode *parent,
       {
 	GIrNodeUnion *union_ = (GIrNodeUnion *)node;
 
-	size = 28;
+	size = 32;
 	size += ALIGN_VALUE (strlen (node->name) + 1, 4);
 	if (union_->gtype_name)
 	  size += ALIGN_VALUE (strlen (union_->gtype_name) + 1, 4);
@@ -1070,6 +1070,92 @@ find_entry (GIrModule   *module,
   return idx;
 }
 
+static GIrNode *
+find_name_in_module (GIrModule   *module,
+		     const gchar *name)
+{
+  GList *l;
+
+  for (l = module->entries; l; l = l->next)
+    {
+      GIrNode *node = (GIrNode *)l->data;
+
+      if (strcmp (node->name, name) == 0)
+	return node;
+    }
+
+  return NULL;
+}
+
+gboolean
+g_ir_find_node (GIrModule  *module,
+		GList      *modules,
+		const char *name,
+		GIrNode   **node_out,
+		GIrModule **module_out)
+{
+  char **names = g_strsplit (name, ".", 0);
+  gint n_names = g_strv_length (names);
+  GIrNode *node = NULL;
+  GList *l;
+
+  if (n_names == 0)
+    {
+      g_warning ("Name can't be empty");
+      goto out;
+    }
+
+  if (n_names > 2)
+    {
+      g_warning ("Too many name parts in '%s'", name);
+      goto out;
+    }
+
+  if (n_names == 1)
+    {
+      *module_out = module;
+      node = find_name_in_module (module, names[0]);
+    }
+  else if (strcmp (names[0], module->name) == 0)
+    {
+      *module_out = module;
+      node = find_name_in_module (module, names[1]);
+    }
+  else
+    {
+      for (l = module->include_modules; l; l = l->next)
+	{
+	  GIrModule *m = l->data;
+
+	  if (strcmp (names[0], m->name) == 0)
+	    {
+	      *module_out = m;
+	      node = find_name_in_module (m, names[1]);
+	      goto out;
+	    }
+	}
+
+      for (l = modules; l; l = l->next)
+	{
+	  GIrModule *m = l->data;
+
+	  if (strcmp (names[0], m->name) == 0)
+	    {
+	      *module_out = m;
+	      node = find_name_in_module (m, names[1]);
+	      goto out;
+	    }
+	}
+    }
+
+ out:
+  g_strfreev (names);
+
+  *node_out = node;
+
+  return node != NULL;
+}
+
 static void
 serialize_type (GIrModule    *module, 
 		GList        *modules,
@@ -1105,7 +1191,8 @@ serialize_type (GIrModule    *module,
   
   if (node->tag < GI_TYPE_TAG_ARRAY)
     {
-      g_string_append_printf (str, "%s", basic[node->tag]);
+      g_string_append_printf (str, "%s%s", basic[node->tag],
+			      node->is_pointer ? "*" : "");
     }
   else if (node->tag == GI_TYPE_TAG_ARRAY)
     {
@@ -1130,14 +1217,19 @@ serialize_type (GIrModule    *module,
 
       iface = find_entry_node (module, modules, node->interface, NULL);
       if (iface)
-	name = iface->name;
+        {
+          if (iface->type == G_IR_NODE_XREF)
+            g_string_append_printf (str, "%s.", ((GIrNodeXRef *)iface)->namespace);
+          name = iface->name;
+        }
       else
 	{
 	  g_warning ("Interface for type reference %s not found", node->interface);
 	  name = node->interface;
 	}
 
-      g_string_append_printf (str, "%s", name);
+      g_string_append_printf (str, "%s%s", name,
+			      node->is_pointer ? "*" : "");
     }
   else if (node->tag == GI_TYPE_TAG_GLIST)
     {
@@ -1268,6 +1360,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	   node->name ? " " : "",
 	   g_ir_node_type_to_string (node->type));
 
+  g_ir_node_compute_offsets (node, module, modules);
+
   switch (node->type)
     {
     case G_IR_NODE_TYPE:
@@ -1278,7 +1372,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	*offset += 4;
 	
 	if (type->tag < GI_TYPE_TAG_ARRAY ||
-	    type->tag == GI_TYPE_TAG_UTF8)
+	    type->tag == GI_TYPE_TAG_UTF8 ||
+	    type->tag == GI_TYPE_TAG_FILENAME)
 	  { 
 	    blob->reserved = 0;
 	    blob->reserved2 = 0;
@@ -1439,7 +1534,10 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->writable = field->writable;
 	blob->reserved = 0;
 	blob->bits = 0;
-	blob->struct_offset = field->offset;
+	if (field->offset >= 0)
+	  blob->struct_offset = field->offset;
+	else
+	  blob->struct_offset = 0xFFFF; /* mark as unknown */
 
         g_ir_node_build_typelib ((GIrNode *)field->type, 
 				 module, modules, strings, types,
@@ -1478,11 +1576,12 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (function->parameters);
 
-	*offset += 16;
-	*offset2 += 8 + n * 12;
+	*offset += sizeof(FunctionBlob);
+	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
 
 	blob->blob_type = BLOB_TYPE_FUNCTION;
 	blob->deprecated = function->deprecated;
+        blob->is_static = !function->is_method;
 	blob->setter = function->is_setter;
 	blob->getter = function->is_getter;
 	blob->constructor = function->is_constructor;
@@ -1530,8 +1629,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (function->parameters);
 
-	*offset += 12;
-	*offset2 += 8 + n * 12;
+	*offset += sizeof(CallbackBlob);
+	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
 
 	blob->blob_type = BLOB_TYPE_CALLBACK;
 	blob->deprecated = function->deprecated;
@@ -1573,8 +1672,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (signal->parameters);
 
-	*offset += 12;
-	*offset2 += 8 + n * 12;
+	*offset += sizeof(SignalBlob);
+	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
 
 	blob->deprecated = signal->deprecated;
 	blob->run_first = signal->run_first;
@@ -1624,8 +1723,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	signature = *offset2;
 	n = g_list_length (vfunc->parameters);
 
-	*offset += 16;
-	*offset2 += 8 + n * 12;
+	*offset += sizeof(VFuncBlob);
+	*offset2 += sizeof(SignatureBlob) + n * sizeof(ArgBlob);
 
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->must_chain_up = 0; /* FIXME */
@@ -1693,6 +1792,9 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->deprecated = struct_->deprecated;
 	blob->reserved = 0;
 	blob->name = write_string (node->name, strings, data, offset2);
+	blob->alignment = struct_->alignment;
+	blob->size = struct_->size;
+
 	if (struct_->gtype_name)
 	  {
 	    blob->unregistered = FALSE;
@@ -1709,7 +1811,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->n_fields = 0;
 	blob->n_methods = 0;
 
-	*offset += 20; 
+	*offset += 24; 
 
 	members = g_list_copy (struct_->members);
 
@@ -1740,11 +1842,13 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->name = write_string (node->name, strings, data, offset2);
 	blob->gtype_name = write_string (boxed->gtype_name, strings, data, offset2);
 	blob->gtype_init = write_string (boxed->gtype_init, strings, data, offset2);
+	blob->alignment = boxed->alignment;
+	blob->size = boxed->size;
 
 	blob->n_fields = 0;
 	blob->n_methods = 0;
 
-	*offset += 20; 
+	*offset += 24; 
 
 	members = g_list_copy (boxed->members);
 
@@ -1772,6 +1876,8 @@ g_ir_node_build_typelib (GIrNode    *node,
 	blob->deprecated = union_->deprecated;
  	blob->reserved = 0;
 	blob->name = write_string (node->name, strings, data, offset2);
+	blob->alignment = union_->alignment;
+	blob->size = union_->size;
 	if (union_->gtype_name)
 	  {
 	    blob->unregistered = FALSE;
@@ -1792,7 +1898,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 
 	if (union_->discriminator_type)
 	  {
-	    *offset += 24;
+	    *offset += 28;
 	    blob->discriminated = TRUE;
 	    g_ir_node_build_typelib ((GIrNode *)union_->discriminator_type, 
 				     module, modules, strings, types,
@@ -1800,7 +1906,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  }
 	else 
 	  {
-	    *offset += 28;
+	    *offset += 32;
 	    blob->discriminated = FALSE;
 	    blob->discriminator_type.offset = 0;
 	  }
@@ -1847,6 +1953,7 @@ g_ir_node_build_typelib (GIrNode    *node,
 	  
 	blob->deprecated = enum_->deprecated;
 	blob->reserved = 0;
+	blob->storage_type = enum_->storage_type;
 	blob->name = write_string (node->name, strings, data, offset2);
 	if (enum_->gtype_name)
 	  {
