@@ -23,6 +23,8 @@
 
 #include <glib.h>
 
+#include "girmodule.h"
+
 G_BEGIN_DECLS
 
 typedef struct _GIrNode GIrNode; 
@@ -226,6 +228,9 @@ struct _GIrNodeInterface
   GList *interfaces;
   GList *prerequisites;
 
+  gint alignment;
+  gint size;
+  
   GList *members;
 };
 
@@ -254,6 +259,7 @@ struct _GIrNodeEnum
   GIrNode node;
 
   gboolean deprecated;
+  gint storage_type;
 
   gchar *gtype_name;
   gchar *gtype_init;
@@ -269,6 +275,9 @@ struct _GIrNodeBoxed
 
   gchar *gtype_name;
   gchar *gtype_init;
+
+  gint alignment;
+  gint size;
   
   GList *members;
 };
@@ -278,9 +287,13 @@ struct _GIrNodeStruct
   GIrNode node;
 
   gboolean deprecated;
+  gboolean disguised;
 
   gchar *gtype_name;
   gchar *gtype_init;
+
+  gint alignment;
+  gint size;
   
   GList *members;
 };
@@ -297,6 +310,9 @@ struct _GIrNodeUnion
   gchar *gtype_name;
   gchar *gtype_init;
 
+  gint alignment;
+  gint size;
+
   gint discriminator_offset;
   GIrNodeType *discriminator_type;
 };
@@ -339,6 +355,19 @@ guint32   write_string              (const gchar *str,
 const gchar * g_ir_node_param_direction_string (GIrNodeParam * node);
 const gchar * g_ir_node_type_to_string         (GIrNodeTypeId type);
 
+gboolean g_ir_find_node (GIrModule  *module,
+			 GList      *modules,
+			 const char *name,
+			 GIrNode   **node_out,
+			 GIrModule **module_out);
+
+/* In giroffsets.c */
+
+void g_ir_node_compute_offsets (GIrNode   *node,
+			        GIrModule *module,
+			        GList     *modules);
+
+
 G_END_DECLS
 
 #endif  /* __G_IR_NODE_H__ */
