@@ -24,6 +24,7 @@
 
 #include <glib.h>
 
+#include "config.h"
 #include "gtypelib.h"
 
 typedef struct {
@@ -53,7 +54,7 @@ validate_interface_blob (ValidateContext *ctx,
 			 guint32        offset,
 			 GError       **error);
 
-DirEntry *
+static DirEntry *
 get_dir_entry_checked (GTypelib *typelib,
 		       guint16    index,
 		       GError   **error)
@@ -151,14 +152,22 @@ g_typelib_check_sanity (void)
 		  n, sizeof (s));					\
       size_check_ok = FALSE; \
     }
+
+  /* When changing the size of a typelib structure, you are required to update
+   * the hardcoded size here.  Do NOT change these to use sizeof(); these
+   * should match whatever is defined in the text specification and serve as
+   * a sanity check on structure modifications.
+   *
+   * Everything else in the code however should be using sizeof().
+   */
   
-  CHECK_SIZE (Header, 104);
+  CHECK_SIZE (Header, 108);
   CHECK_SIZE (DirEntry, 12);
   CHECK_SIZE (SimpleTypeBlob, 4);
   CHECK_SIZE (ArgBlob, 12);
   CHECK_SIZE (SignatureBlob, 8);
   CHECK_SIZE (CommonBlob, 8);
-  CHECK_SIZE (FunctionBlob, 16);
+  CHECK_SIZE (FunctionBlob, 20);
   CHECK_SIZE (InterfaceTypeBlob, 4);
   CHECK_SIZE (ArrayTypeBlob, 8);
   CHECK_SIZE (ParamTypeBlob, 4);
@@ -167,7 +176,7 @@ g_typelib_check_sanity (void)
   CHECK_SIZE (ValueBlob, 12);
   CHECK_SIZE (FieldBlob, 12);
   CHECK_SIZE (RegisteredTypeBlob, 16);
-  CHECK_SIZE (StructBlob, 20);
+  CHECK_SIZE (StructBlob, 24);
   CHECK_SIZE (EnumBlob, 20);
   CHECK_SIZE (PropertyBlob, 12);
   CHECK_SIZE (SignalBlob, 12);
@@ -176,7 +185,7 @@ g_typelib_check_sanity (void)
   CHECK_SIZE (InterfaceBlob, 28);
   CHECK_SIZE (ConstantBlob, 20);
   CHECK_SIZE (AnnotationBlob, 12);
-  CHECK_SIZE (UnionBlob, 28);
+  CHECK_SIZE (UnionBlob, 32);
 #undef CHECK_SIZE
 
   g_assert (size_check_ok);
@@ -306,7 +315,7 @@ validate_header (ValidateContext  *ctx,
     }
 
   if (header->entry_blob_size != 12 ||
-      header->function_blob_size != 16 ||
+      header->function_blob_size != 20 ||
       header->callback_blob_size != 12 ||
       header->signal_blob_size != 12 ||
       header->vfunc_blob_size != 16 ||
@@ -319,10 +328,10 @@ validate_header (ValidateContext  *ctx,
       header->annotation_blob_size != 12 ||
       header->signature_blob_size != 8 ||
       header->enum_blob_size != 20 ||
-      header->struct_blob_size != 20 ||
+      header->struct_blob_size != 24 ||
       header->object_blob_size != 32 ||
       header->interface_blob_size != 28 ||
-      header->union_blob_size != 28)
+      header->union_blob_size != 32)
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
@@ -704,6 +713,7 @@ validate_function_blob (ValidateContext *ctx,
   GTypelib *typelib = ctx->typelib;
   FunctionBlob *blob;
   SignatureBlob *sigblob;
+  gboolean is_method;
 
   if (typelib->len < offset + sizeof (FunctionBlob))
     {
@@ -721,7 +731,7 @@ validate_function_blob (ValidateContext *ctx,
       g_set_error (error,
 		   G_TYPELIB_ERROR,
 		   G_TYPELIB_ERROR_INVALID_BLOB,
-		   "Wrong blob type");
+		   "Wrong blob type %d, expected function", blob->blob_type);
       return FALSE;
     }
 
@@ -732,12 +742,27 @@ validate_function_blob (ValidateContext *ctx,
   
   if (!validate_name (typelib, "function symbol", typelib->data, blob->symbol, error))
     return FALSE; 
+
+  switch (container_type)
+    {
+    case BLOB_TYPE_BOXED:
+    case BLOB_TYPE_STRUCT:
+    case BLOB_TYPE_UNION:
+    case BLOB_TYPE_OBJECT:
+    case BLOB_TYPE_INTERFACE:
+      is_method = !(blob->constructor || blob->setter || blob->getter || blob->wraps_vfunc);
+      break;
+    default:
+      is_method = FALSE;
+    }
   
   if (blob->constructor)
     {
       switch (container_type)
 	{
 	case BLOB_TYPE_BOXED:
+	case BLOB_TYPE_STRUCT:
+	case BLOB_TYPE_UNION:
 	case BLOB_TYPE_OBJECT:
 	case BLOB_TYPE_INTERFACE:
 	  break;
@@ -779,7 +804,6 @@ validate_function_blob (ValidateContext *ctx,
     }
 
   /* FIXME: validate index range */
-  /* FIXME: validate "this" argument for methods */
 
   if (!validate_signature_blob (typelib, blob->signature, error))
     return FALSE;
@@ -792,7 +816,6 @@ validate_function_blob (ValidateContext *ctx,
 							   blob->signature,
 							   error);
       InterfaceTypeBlob *iface_type;
-      InterfaceBlob *iface;
 
       if (!simple)
 	return FALSE;
@@ -861,14 +884,36 @@ validate_constant_blob (GTypelib     *typelib,
 			guint32        offset,
 			GError       **error)
 {
-  gint value_size[] = {
-    0, 4, 1, 1, 2, 2, 4, 4, 8, 8, 
-    sizeof (gint), sizeof (guint), 
-    sizeof (glong), sizeof (gulong),
-    sizeof (gssize), sizeof (gsize),
-    sizeof (gfloat), sizeof (gdouble), 
-    0, 0
-  }; 
+  guint value_size[] = {
+    0, /* VOID */
+    4, /* BOOLEAN */
+    1, /* INT8 */
+    1, /* UINT8 */
+    2, /* INT16 */
+    2, /* UINT16 */
+    4, /* INT32 */
+    4, /* UINT32 */
+    8, /* INT64 */
+    8, /* UINT64 */
+    sizeof (gint),
+    sizeof (guint),
+    sizeof (glong),
+    sizeof (gulong),
+    sizeof (gssize),
+    sizeof (gsize),
+    sizeof (gfloat),
+    sizeof (gdouble),
+    sizeof (time_t),
+    0, /* GTYPE */
+    0, /* UTF8 */
+    0, /* FILENAME */
+    0, /* ARRAY */
+    0, /* INTERFACE */
+    0, /* GLIST */
+    0, /* GSLIST */
+    0, /* GHASH */
+    0, /* ERROR */
+  };
   ConstantBlob *blob;
   SimpleTypeBlob *type;
 
@@ -909,7 +954,7 @@ validate_constant_blob (GTypelib     *typelib,
     }
   
   type = (SimpleTypeBlob *)&typelib->data[offset + G_STRUCT_OFFSET (ConstantBlob, type)];
-  if (type->reserved == 0)
+  if (type->reserved == 0 && type->reserved2 == 0)
     {
       if (type->tag == 0)
 	{
@@ -1173,23 +1218,13 @@ validate_struct_blob (ValidateContext *ctx,
 		   "Wrong blob type");
       return FALSE;
     }
-  
-  if ((blob->blob_type == BLOB_TYPE_BOXED && blob->unregistered) ||
-      (blob->blob_type == BLOB_TYPE_STRUCT && !blob->unregistered))
-    {
-      g_set_error (error,
-		   G_TYPELIB_ERROR,
-		   G_TYPELIB_ERROR_INVALID_BLOB,
-		   "Registration/blob type mismatch");
-      return FALSE;
-    }
 
   if (!validate_name (typelib, "struct", typelib->data, blob->name, error))
     return FALSE; 
 
   push_context (ctx, get_string_nofail (typelib, blob->name));
   
-  if (blob_type == BLOB_TYPE_BOXED)
+  if (!blob->unregistered)
     {
       if (!validate_name (typelib, "boxed", typelib->data, blob->gtype_name, error))
 	return FALSE; 
@@ -1253,8 +1288,7 @@ validate_enum_blob (ValidateContext *ctx,
 {
   GTypelib *typelib = ctx->typelib;
   EnumBlob *blob;
-  ValueBlob *v1, *v2;
-  gint i, j; 
+  gint i;
 
   if (typelib->len < offset + sizeof (EnumBlob))
     {
@@ -1880,44 +1914,6 @@ g_typelib_error_quark (void)
   return quark;
 }
 
-static const char*
-find_some_symbol (GTypelib *typelib)
-{
-  Header *header = (Header *) typelib->data;
-  gint i;
-
-  for (i = 0; i < header->n_entries; i++)
-    {
-      DirEntry *entry;
-      
-      entry = g_typelib_get_dir_entry (typelib, i + 1);
-
-      switch (entry->blob_type)
-        {
-        case BLOB_TYPE_FUNCTION:
-          {
-            FunctionBlob *blob = (FunctionBlob *) &typelib->data[entry->offset];
-            
-            if (blob->symbol)
-              return g_typelib_get_string (typelib, blob->symbol);
-          }
-          break;
-        case BLOB_TYPE_OBJECT:
-          {
-            RegisteredTypeBlob *blob = (RegisteredTypeBlob *) &typelib->data[entry->offset];
-            
-            if (blob->gtype_init)
-              return g_typelib_get_string (typelib, blob->gtype_init);
-          }
-          break;
-        default:
-          break;
-        }
-    }
-
-  return NULL;
-}
-
 static inline void
 _g_typelib_init (GTypelib *typelib)
 {
@@ -1926,76 +1922,75 @@ _g_typelib_init (GTypelib *typelib)
   header = (Header *) typelib->data;
   if (header->shared_library)
     {
-      const gchar *shlib;
+      const gchar *shlib_str;
+      GModule *app_module = NULL;
 
-      shlib = g_typelib_get_string (typelib, header->shared_library);
+      shlib_str = g_typelib_get_string (typelib, header->shared_library);
       /* note that NULL shlib means to open the main app, which is allowed */
 
-      /* If we do have a shared lib, first be sure the main app isn't already linked to it */
-      if (shlib != NULL)
+      if (shlib_str != NULL)
         {
-          const char *symbol_in_module;
-           
-          symbol_in_module = find_some_symbol (typelib);
-          if (symbol_in_module != NULL)
+          gchar **shlibs;
+          gint i;
+
+          /* shared-library is a comma-separated list of libraries */
+          shlibs = g_strsplit (shlib_str, ",", 0);
+
+          /* We load all passed libs unconditionally as if the same library is loaded
+           * again with dlopen(), the same file handle will be returned. See bug:
+           * http://bugzilla.gnome.org/show_bug.cgi?id=555294
+           */
+          for (i = 0; shlibs[i]; i++)
             {
-              typelib->module = g_module_open (NULL, G_MODULE_BIND_LAZY);
-              if (typelib->module == NULL)
+              GModule *module;
+
+              /* Glade's autoconnect feature and OpenGL's extension mechanism
+               * as used by Clutter rely on dlopen(NULL) to work as a means of
+               * accessing the app's symbols. This keeps us from using
+               * G_MODULE_BIND_LOCAL. BIND_LOCAL may have other issues as well;
+               * in general libraries are not expecting multiple copies of
+               * themselves and are not expecting to be unloaded. So we just
+               * load modules globally for now.
+               */
+
+	      module = g_module_open (shlibs[i], G_MODULE_BIND_LAZY);
+
+	      if (module == NULL)
+	        {
+	          GString *shlib_full = g_string_new (shlibs[i]);
+
+	          /* Prefix with "lib", try both .la and .so */
+	          if (!g_str_has_prefix (shlib_full->str, "lib"))
+	            g_string_prepend (shlib_full, "lib");
+	          g_string_append (shlib_full, ".la");
+	          module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
+	          if (module == NULL)
+	            g_string_overwrite (shlib_full, strlen (shlib_full->str)-2, SHLIB_SUFFIX);
+	          module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
+
+	          g_string_free (shlib_full, TRUE);
+	        }
+
+	      if (module == NULL)
                 {
-                  g_warning ("Could not open main app as GModule: %s",
-                             g_module_error ());
+                  g_warning ("Failed to load shared library '%s' referenced by the typelib: %s",
+                             shlibs[i], g_module_error ());
                 }
               else
                 {
-                  void *sym;
-                  if (!g_module_symbol (typelib->module, symbol_in_module, &sym))
-                    {
-                      /* we will try opening the shlib, symbol is not in app already */
-                      g_module_close (typelib->module);
-                      typelib->module = NULL;
-                    }
+                  typelib->modules = g_list_append (typelib->modules, module);
                 }
-            }
-          else
-            {
-              g_warning ("Could not find any symbols in typelib");
-            }
+          }
+
+          g_strfreev (shlibs);
         }
-     
-      if (typelib->module == NULL && shlib != NULL)
-        {
-	  GString *shlib_full;
 
-          /* Glade's autoconnect feature and OpenGL's extension mechanism
-           * as used by Clutter rely on dlopen(NULL) to work as a means of
-           * accessing the app's symbols. This keeps us from using
-           * G_MODULE_BIND_LOCAL. BIND_LOCAL may have other issues as well;
-           * in general libraries are not expecting multiple copies of
-           * themselves and are not expecting to be unloaded. So we just
-           * load modules globally for now.
-           */
-
-	  typelib->module = g_module_open (shlib, G_MODULE_BIND_LAZY);
-
-	  if (typelib->module == NULL)
-	    {
-	      shlib_full = g_string_new (shlib);
-
-	      /* Prefix with "lib", try both .la and .so */
-	      if (!g_str_has_prefix (shlib_full->str, "lib"))
-		g_string_prepend (shlib_full, "lib");
-	      g_string_append (shlib_full, ".la");
-	      typelib->module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
-	      if (typelib->module == NULL)
-		g_string_overwrite (shlib_full, strlen (shlib_full->str)-2, G_MODULE_SUFFIX);
-	      typelib->module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
-
-	      g_string_free (shlib_full, TRUE);
-	    }
-	  if (typelib->module == NULL)
-            g_warning ("Failed to load shared library '%s' referenced by the typelib: %s",
-                       shlib, g_module_error ());
-        }
+        /* we should make sure the app_module in the end of list so that
+         * it's last symbol source when loading any symbols from modules.
+         * See comments in g_typelib_symbol */
+        app_module = g_module_open (NULL, G_MODULE_BIND_LAZY);
+        if (app_module)
+          typelib->modules = g_list_append (typelib->modules, app_module);
     }
 }
 
@@ -2019,6 +2014,7 @@ g_typelib_new_from_memory (guchar *memory, gsize len)
   meta->data = memory;
   meta->len = len;
   meta->owns_memory = TRUE;
+  meta->modules = NULL;
   _g_typelib_init (meta);
   return meta;
 }
@@ -2041,6 +2037,7 @@ g_typelib_new_from_const_memory (const guchar *memory, gsize len)
   meta->data = (guchar *) memory;
   meta->len = len;
   meta->owns_memory = FALSE;
+  meta->modules = NULL;
   _g_typelib_init (meta);
   return meta;
 }
@@ -2081,28 +2078,56 @@ g_typelib_free (GTypelib *typelib)
   else
     if (typelib->owns_memory)
       g_free (typelib->data);
-  if (typelib->module)
-    g_module_close (typelib->module);
+  if (typelib->modules)
+    {
+      g_list_foreach (typelib->modules, (GFunc) g_module_close, NULL);
+      g_list_free (typelib->modules);
+    }
   g_free (typelib);
 }
 
-/**
- * g_typelib_set_module:
- * @typelib: a #GTypelib instance
- * @module: a #GModule; takes ownership of this module
- * 
- * Sets the target module for all symbols referenced by the typelib.
- **/
-void
-g_typelib_set_module (GTypelib *typelib, GModule *module)
-{
-  if (typelib->module)
-    g_module_close (typelib->module);
-  typelib->module = module;
-}
-
 const gchar *
-g_typelib_get_namespace(GTypelib *typelib)
+g_typelib_get_namespace (GTypelib *typelib)
 {
   return g_typelib_get_string (typelib, ((Header *) typelib->data)->namespace);
 }
+
+/**
+ * g_typelib_symbol:
+ * @symbol_name: name of symbol to be loaded
+ * @symbol: returns a pointer to the symbol value
+ *
+ * Loads a symbol from #GTypelib.
+ *
+ * Return value: #TRUE on success
+ **/
+gboolean
+g_typelib_symbol (GTypelib *typelib, const char *symbol_name, gpointer *symbol)
+{
+  GList *l;
+
+  /*
+   * We want to be able to add symbols to an app or an auxiliary
+   * library to fill in gaps in an introspected library. However,
+   * normally we would only look for symbols in the main library
+   * (the first items in typelib->modules).
+   *
+   * A more elaborate solution is probably possible, but as a
+   * simple approach for now, if we fail to find a symbol we look
+   * for it in the global module (the last item in type->modules).
+   *
+   * This would not be very efficient if it happened often, since
+   * we always do the failed lookup above first, but very few
+   * symbols should be outside of the main libraries in
+   * typelib->modules so it doesn't matter.
+   */
+  for (l = typelib->modules; l; l = l->next)
+    {
+      GModule *module = l->data;
+
+      if (g_module_symbol (module, symbol_name, symbol))
+        return TRUE;
+    }
+
+  return FALSE;
+}
