@@ -2,20 +2,20 @@
 # GObject-Introspection - a framework for introspecting GObject libraries
 # Copyright (C) 2008  Johan Dahlin
 #
-# This program is free software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 2
-# of the License, or (at your option) any later version.
+# This library is free software; you can redistribute it and/or
+# modify it under the terms of the GNU Lesser General Public
+# License as published by the Free Software Foundation; either
+# version 2 of the License, or (at your option) any later version.
 #
-# This program is distributed in the hope that it will be useful,
+# This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# Lesser General Public License for more details.
 #
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-# 02110-1301, USA.
+# You should have received a copy of the GNU Lesser General Public
+# License along with this library; if not, write to the
+# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+# Boston, MA 02111-1307, USA.
 #
 
 import os
@@ -142,7 +142,8 @@ class Transformer(object):
                 self._names.aliases[node.name] = (nsname, node)
             elif isinstance(node, (GLibBoxed, Interface, Class)):
                 self._names.type_names[node.type_name] = (nsname, node)
-            self._names.names[node.name] = (nsname, node)
+            giname = '%s.%s' % (nsname, node.name)
+            self._names.names[giname] = (nsname, node)
             if hasattr(node, 'ctype'):
                 self._names.ctypes[node.ctype] = (nsname, node)
             elif hasattr(node, 'symbol'):
@@ -248,7 +249,7 @@ class Transformer(object):
         raise ValueError("Unmatched length parameter name %r"\
                              % (target_name, ))
 
-    def _pair_annotations(self, params):
+    def _pair_annotations(self, params, return_):
         names = {}
         for param in params:
             if param.name in names:
@@ -258,6 +259,9 @@ class Transformer(object):
             if isinstance(param.type, Array):
                 self._pair_array(params, param)
 
+        if isinstance(return_.type, Array):
+            self._pair_array(params, return_)
+
     # We take the annotations from the parser as strings; here we
     # want to split them into components, so:
     # (transfer full) -> {'transfer' : [ 'full' ]}
@@ -274,9 +278,9 @@ class Transformer(object):
         directives = symbol.directives()
         parameters = list(self._create_parameters(
             symbol.base_type, directives))
-        self._pair_annotations(parameters)
         return_ = self._create_return(symbol.base_type.base_type,
                                       directives.get('return', {}))
+        self._pair_annotations(parameters, return_)
         name = self._strip_namespace_func(symbol.ident)
         func = Function(name, return_, parameters, symbol.ident)
         self._parse_deprecated(func, directives)
@@ -331,7 +335,9 @@ class Transformer(object):
                 if child_list:
                     size_opt = 'fixed-size=%d' % (child_list[0].const_int, )
                     opts['array'].append(size_opt)
-            ftype = self._create_type(symbol.base_type, opts, True)
+            ftype = self._create_type(symbol.base_type, opts,
+                                      is_param=False, is_retval=False)
+            ftype = self.resolve_param_type(ftype)
             # Fields are assumed to be read-write
             # (except for Objects, see also glibtransformer.py)
             node = Field(symbol.ident, ftype, symbol.ident,
@@ -343,6 +349,9 @@ class Transformer(object):
         if (ctype == CTYPE_POINTER and
             symbol.base_type.base_type.type == CTYPE_FUNCTION):
             node = self._create_callback(symbol)
+        elif (ctype == CTYPE_POINTER and
+            symbol.base_type.base_type.type == CTYPE_STRUCT):
+            node = self._create_typedef_struct(symbol, disguised=True)
         elif ctype == CTYPE_STRUCT:
             node = self._create_typedef_struct(symbol)
         elif ctype == CTYPE_UNION:
@@ -366,22 +375,49 @@ class Transformer(object):
                 "symbol %r of type %s" % (symbol.ident, ctype_name(ctype)))
         return node
 
-    def _parse_ctype(self, ctype):
+    def _canonicalize_ctype(self, ctype):
         # First look up the ctype including any pointers;
         # a few type names like 'char*' have their own aliases
         # and we need pointer information for those.
         firstpass = type_name_from_ctype(ctype)
 
+        # If we have a particular alias for this, skip deep
+        # canonicalization to prevent changing
+        # e.g. char* -> int8*
+        if firstpass != ctype:
+            return firstpass
+
+        # We're also done if the type is already a fundamental
+        # known type, or there are no pointers.
+        if ctype in type_names or not firstpass.endswith('*'):
+            return firstpass
+
+        # We have a pointer type.
+        # Strip the end pointer, canonicalize our base type
+        base = firstpass[:-1]
+        canonical_base = self._canonicalize_ctype(base)
+
+        # Append the pointer again
+        canonical = canonical_base + '*'
+
+        return canonical
+
+    def parse_ctype(self, ctype, is_member=False):
+        canonical = self._canonicalize_ctype(ctype)
+
         # Remove all pointers - we require standard calling
         # conventions.  For example, an 'int' is always passed by
         # value (unless it's out or inout).
-        derefed = firstpass.replace('*', '')
+        derefed_typename = canonical.replace('*', '')
 
-        # Canonicalize our type again, this time without the pointer;
-        # this ensures we turn e.g. plain "guint" => "int"
-        return type_name_from_ctype(derefed)
+        # Preserve "pointerness" of struct/union members
+        if (is_member and canonical.endswith('*') and
+            derefed_typename in BASIC_GIR_TYPES):
+            return 'any'
+        else:
+            return derefed_typename
 
-    def _create_type(self, source_type, options, is_param):
+    def _create_type(self, source_type, options, is_param, is_retval):
         ctype = self._create_source_type(source_type)
         if ctype == 'va_list':
             raise SkipError()
@@ -390,42 +426,52 @@ class Transformer(object):
         elif ctype == 'FILE*':
             raise SkipError
 
+        canonical_ctype = self._canonicalize_ctype(ctype)
+
         # Now check for a list/map/array type
-        if ctype in self._list_ctypes:
+        if canonical_ctype in self._list_ctypes:
             param = options.get('element-type')
             if param:
-                contained_type = self._parse_ctype(param[0])
+                contained_type = self.parse_ctype(param[0])
             else:
                 contained_type = None
-            derefed_name = self._parse_ctype(ctype)
+            derefed_name = self.parse_ctype(ctype)
             rettype = List(derefed_name,
                            ctype,
                            contained_type)
-        elif ctype in self._map_ctypes:
+        elif canonical_ctype in self._map_ctypes:
             param = options.get('element-type')
             if param:
-                key_type = self._parse_ctype(param[0])
-                value_type = self._parse_ctype(param[1])
+                key_type = self.parse_ctype(param[0])
+                value_type = self.parse_ctype(param[1])
             else:
                 key_type = None
                 value_type = None
-            derefed_name = self._parse_ctype(ctype)
+            derefed_name = self.parse_ctype(ctype)
             rettype = Map(derefed_name,
                           ctype,
                           key_type, value_type)
-        elif (ctype in default_array_types) or ('array' in options):
-            derefed_name = ctype[:-1] if ctype[-1] == '*' else ctype
+        elif ((is_param and canonical_ctype in default_array_types)
+              or ('array' in options)):
+            if canonical_ctype[-1] == '*':
+                derefed_name = canonical_ctype[:-1]
+            else:
+                derefed_name = canonical_ctype
             rettype = Array(ctype,
-                            self._parse_ctype(derefed_name))
+                            self.parse_ctype(derefed_name))
             array_opts = dict([opt.split('=')
                                for opt in options.get('array', [])])
             if 'length' in array_opts:
                 rettype.length_param_name = array_opts['length']
+                rettype.zeroterminated = False
             if 'fixed-size' in array_opts:
                 rettype.size = array_opts['fixed-size']
                 rettype.zeroterminated = False
+            if 'zero-terminated' in array_opts:
+                rettype.zeroterminated = array_opts['zero-terminated'] != '0'
         else:
-            derefed_name = self._parse_ctype(ctype)
+            derefed_name = self.parse_ctype(ctype,
+                                            not (is_param or is_retval))
             rettype = Type(derefed_name, ctype)
 
         # Deduce direction for some types passed by reference that
@@ -513,7 +559,9 @@ class Transformer(object):
             if 'transfer' not in options:
                 options['transfer'] = ['none']
         else:
-            ptype = self._create_type(symbol.base_type, options, True)
+            ptype = self._create_type(symbol.base_type, options,
+                                      is_param=True, is_retval=False)
+            ptype = self.resolve_param_type(ptype)
         param = Parameter(symbol.ident, ptype)
         for option, data in options.iteritems():
             if option in ['in-out', 'inout']:
@@ -541,7 +589,8 @@ class Transformer(object):
             options_map = {}
         else:
             options_map = self._parse_options(options)
-        rtype = self._create_type(source_type, options_map, False)
+        rtype = self._create_type(source_type, options_map,
+                                  is_param=False, is_retval=True)
         rtype = self.resolve_param_type(rtype)
         return_ = Return(rtype)
         self._handle_generic_param_options(return_, options_map)
@@ -549,6 +598,8 @@ class Transformer(object):
             if option in ('transfer', 'transfer-inferred',
                           'element-type', 'out'):
                 pass
+            elif option.startswith(('element-type', 'array')):
+                pass
             else:
                 print 'Unhandled return type annotation option: %r' % (
                     option, )
@@ -567,9 +618,9 @@ class Transformer(object):
         const = Constant(name, type_name, value)
         return const
 
-    def _create_typedef_struct(self, symbol):
+    def _create_typedef_struct(self, symbol, disguised=False):
         name = self.remove_prefix(symbol.ident)
-        struct = Struct(name, symbol.ident)
+        struct = Struct(name, symbol.ident, disguised)
         self._typedefs_ns[symbol.ident] = struct
         self._create_struct(symbol)
         return struct
@@ -650,6 +701,12 @@ class Transformer(object):
             return type_names[type_name]
         except KeyError, e:
             pass
+
+        if ctype:
+            ctype = ctype.replace('*', '')
+            resolved = names.ctypes.get(ctype)
+            if resolved:
+                return self._typepair_to_str(resolved)
         type_name = self.remove_prefix(type_name)
         resolved = names.aliases.get(type_name)
         if resolved:
@@ -657,11 +714,6 @@ class Transformer(object):
         resolved = names.names.get(type_name)
         if resolved:
             return self._typepair_to_str(resolved)
-        if ctype:
-            ctype = ctype.replace('*', '')
-            resolved = names.ctypes.get(ctype)
-            if resolved:
-                return self._typepair_to_str(resolved)
         resolved = names.type_names.get(type_name)
         if resolved:
             return self._typepair_to_str(resolved)
